#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <iostream.h>
#include <sys/neutrino.h>
#include <sys/syspage.h> 
#include <sys/netmgr.h>
#include <pthread.h>
#include <signal.h>
#include <algorithm>

static void out( char s ) {
   int policy;
   sched_param param;
   pthread_getschedparam( pthread_self(), &policy, &param );
   cout << s << param.sched_curpriority << flush;
};

// эталонная функция, которая запускается периодически
static void mon1( void ) { out( '.' ); };
static void mon2( void ) { out( '*' ); };  
static void mon3( void ) { out( '+' ); };  

inline uint64_t cycles2nsec( uint64_t c ) {
   const static uint64_t cps = SYSPAGE_ENTRY( qtime )->cycles_per_sec;  // частота процессора:
   return ( c * 1000000000 ) / cps;
};     

struct timestat {
private:
   uint64_t               prev;
public:
   uint64_t               num;
   double                mean, disp, tmin, tmax;
   timestat( void ) {
      mean = disp = tmin = tmax = 0.0;
      num = 0;
   };
   void operator ++( void ) {
      uint64_t next = ClockCycles(), delta;
      if( num != 0 ) {
         double delta = cycles2nsec( next - prev );
         if( num == 1 ) tmin = tmax = delta;
         else tmin = min( tmin, delta ), tmax = max( tmax, delta );
         mean += delta;
         disp += delta * delta;
      };
      prev = next;
      num++;
   };
   void operator !( void ) {
      mean /= ( num - 1 );
      disp = sqrt( disp / ( num - 1 ) - mean * mean );
   };
};

void* syncthread( void * );

class thrblock {
private:
   static int            code;
   bool                  ok, st;
public:
   pthread_t           tid;
   struct sigevent  event;
   timer_t                timer;
   int                      chid;
   void* ( *func )( void* );
   sched_param param;
   timestat              sync;
   thrblock( void ( *dofunc )( void ), unsigned long millisec, unsigned short priority, bool statist = false ) {
      // канал, к которому будет подключаться таймер для для интервальных сообщений
      if( !( ok = ( ( chid = ChannelCreate( 0 ) ) >= 0 ) ) ) return;
      // создание синхронизирующего канала для потока выполнения
      event.sigev_coid = ConnectAttach( ND_LOCAL_NODE, 0, chid, _NTO_SIDE_CHANNEL, 0 );
      if( !( ok = ( event.sigev_coid >= 0 ) ) ) return;
      // синхронизируемая функция
      func = (void* (*)( void* ) ) dofunc;
      int policy;
      pthread_getschedparam( pthread_self(), &policy, &param );
      st = statist;
      event.sigev_code = code++;
      event.sigev_notify = SIGEV_PULSE;
      event.sigev_priority = priority;
      // создание периодического таймера
      if( !( ok = ( timer_create( CLOCK_REALTIME, &event, &timer ) == 0 ) ) ) return;
      // запуск синхронизирующего потока
      if( !( ok = ( pthread_create( &tid, NULL, &syncthread, (void*)this ) == EOK ) ) ) return;
      // установка интервала и запуск таймера для синхронизатора
      struct itimerspec itime;
      nsec2timespec( &itime.it_value, millisec * 1000000ull ); 
      itime.it_interval =  itime.it_value;
      if( !( ok = ( timer_settime( timer, 0, &itime, NULL ) == 0 ) ) ) return;
   };
    bool OK( void ) { return ok; };
    bool statistic( void ) { return st; }; 
};

int thrblock::code = _PULSE_CODE_MINAVAIL; 

void* syncthread( void *block ) {
   thrblock *p = (thrblock*)block;
   struct _pulse buf;
   pthread_attr_t attr;
   while( true ) {
      // ожидание пульса синхронизации
      MsgReceivePulse( p->chid, &buf, sizeof( struct _pulse ), NULL );
      pthread_attr_init( &attr );
      pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
//      pthread_attr_setinheritsched( &attr, PTHREAD_EXPLICIT_SCHED );
//      pthread_attr_setschedparam( &attr, &p->param );
      pthread_create( NULL, &attr, p->func, NULL );
      if( p->statistic() ) ++p->sync;
   };
};

// "пустой" обработчик сигнала (SIGINT - ^C)
inline static void empty( int signo ) { };

int main( int argc, char **argv ) {
   // с этой точки реакция на ^C по умолчанию  - блокируется
   signal( SIGINT, empty ); 
   void ( *funcs[] )( void ) = { &mon1, &mon2, &mon3 };
   int period[] = { 317, 171, 77 };
   int proirity[] = { 15, 5, 25 };
   int num = sizeof( funcs ) / sizeof( *funcs );
   // запуск синхронизированных потоков
   thrblock** tb = new (thrblock*)[ num ]; 
   for( int i = 0; i < num; i++ ) {
      tb[ i ] = new thrblock( funcs[ i ], period[ i ], proirity[ i ], true );
      if( !tb[ i ]->OK() ) perror( "synchro thread create" ), exit( EXIT_FAILURE );
   };
   pause();
   cout << endl << "Monitoring finalisation!" << endl;
   const double n2m = 1000000.;
   for( int i = 0; i < num; i++ ) {
      timestat *p = &tb[ i ]->sync;
      !( *p );              // финализация статистики
      cout << i << '\t' << p->num << "\t=> "  << p->mean / n2m
              << " [" << p->tmin / n2m << "..." << p->tmax / n2m << "]\t~" << p->disp / n2m
              << " (" << p->disp / p->mean * 100. << "%)" << endl;
   };
   return EXIT_SUCCESS;
};

